// Flutter Packages
import 'package:flutter/material.dart';

// Dart Packages
import 'dart:async';

// Firebase Packages
import 'package:cloud_firestore/cloud_firestore.dart';

// Data Models
import 'models/view_type.dart';
import 'models/wrap_options.dart';

// Widgets
import 'widgets/defaults/bottom_loader.dart';
import 'widgets/defaults/empty_screen.dart';
import 'widgets/defaults/initial_loader.dart';
import 'widgets/views/build_pagination.dart';

class FirestorePagination extends StatefulWidget {
  const FirestorePagination({
    Key? key,
    required this.query,
    required this.itemBuilder,
    this.limit = 10,
    this.viewType = ViewType.list,
    this.isRealtime = true,
    this.gridDelegate = const SliverGridDelegateWithFixedCrossAxisCount(
      crossAxisCount: 2,
      childAspectRatio: 1,
    ),
    this.wrapOptions = const WrapOptions(),
    this.onEmpty = const EmptyScreen(),
    this.bottomLoader = const BottomLoader(),
    this.initialLoader = const InitialLoader(),
  }) : super(key: key);

  final Query query;
  final Widget Function(BuildContext, DocumentSnapshot, int) itemBuilder;
  final int limit;
  final ViewType viewType;
  final bool isRealtime;
  final SliverGridDelegate gridDelegate;
  final WrapOptions wrapOptions;
  final Widget onEmpty;
  final Widget bottomLoader;
  final Widget initialLoader;

  @override
  State<FirestorePagination> createState() => _FirestorePaginationState();
}

class _FirestorePaginationState extends State<FirestorePagination> {
  final List<DocumentSnapshot> _docs = [];
  StreamSubscription<QuerySnapshot>? _streamSub;
  StreamSubscription<QuerySnapshot>? _realtimeStreamSub;
  final ScrollController _controller = ScrollController();
  bool _isInitialLoading = true;
  bool _isFetching = false;
  bool _isEnded = false;

  Future<void> _loadDocuments({bool getMore = true}) async {
    final tempSub = _streamSub;

    if (getMore) setState(() => _isFetching = true);

    int docsLimit = _docs.length + (getMore ? widget.limit : 0);
    Query docsQuery = widget.query.limit(docsLimit);
    if (_docs.isNotEmpty) {
      docsQuery = docsQuery.startAtDocument(_docs.first);
    }
    _streamSub = docsQuery.snapshots().listen((QuerySnapshot snapshot) async {
      await tempSub?.cancel();

      _docs.clear();
      _docs.addAll(snapshot.docs);

      bool isDocRemoved = snapshot.docChanges.any(
        (DocumentChange change) => change.type == DocumentChangeType.removed,
      );

      _isFetching = false;
      if (!_isEnded && !isDocRemoved) {
        _isEnded = snapshot.docs.length < docsLimit;
      }

      if (isDocRemoved || _isInitialLoading) {
        _isInitialLoading = false;
        if (snapshot.docs.isNotEmpty) {
          await _loadDocuments(getMore: false);
        } else {
          _streamSub?.cancel();
        }
        if (widget.isRealtime) _setRealtimeListner();
      }

      if (mounted) setState(() {});
    });
  }

  Future<void> _setRealtimeListner() async {
    final tempSub = _realtimeStreamSub;

    Query latestDocQuery = widget.query.limit(1);
    if (_docs.isNotEmpty) {
      latestDocQuery = latestDocQuery.endBeforeDocument(_docs.first);
    }

    _realtimeStreamSub = latestDocQuery.snapshots().listen(
      (QuerySnapshot snapshot) async {
        await tempSub?.cancel();
        if (snapshot.docs.isEmpty ||
            snapshot.docs.first.metadata.hasPendingWrites) return;

        _docs.insert(0, snapshot.docs.first);
        await _setRealtimeListner();
        _loadDocuments(getMore: false);
      },
    );
  }

  void _scrollListener() {
    final ScrollPosition position = _controller.position;
    if (position.pixels >= (position.maxScrollExtent - 50) &&
        !_isInitialLoading &&
        !_isFetching &&
        !_isEnded) {
      _loadDocuments();
    }
  }

  @override
  void initState() {
    super.initState();
    _loadDocuments();

    _controller.addListener(_scrollListener);
  }

  @override
  void dispose() {
    _streamSub?.cancel();
    _realtimeStreamSub?.cancel();
    _controller.removeListener(_scrollListener);
    _controller.dispose();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return _isInitialLoading
        ? widget.initialLoader
        : _docs.isEmpty
            ? widget.onEmpty
            : BuildPagination(
                controller: _controller,
                items: _docs,
                itemBuilder: widget.itemBuilder,
                isLoading: _isFetching,
                bottomLoader: widget.bottomLoader,
                gridDelegate: widget.gridDelegate,
                viewType: widget.viewType,
                wrapOptions: widget.wrapOptions,
              );
  }
}
